/*
 * Copyright (C) 2015 Assured Information Security, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "qemu/osdep.h"
#ifdef CONFIG_SYSLOG_LOGGING
# include "logging.h"
#else
# include "stdio.h"
#endif

#include "sysemu/sysemu.h"
#include "hw/hw.h"
#include "hw/i386/pc.h"
#include "hw/acpi/acpi.h"
#include "qemu/config-file.h"
#include "qapi/qmp-event.h"
#include "hw/xen/xen.h"
#include "hw/xen/xen_backend.h"

/**
 * ACPI_LOG: information to always log (errors & important low-volume events)
 * @param fmt,... printf style arguments
 */
#define ACPI_LOG(fmt, ...)                                                   \
    do {                                                                   \
            fprintf(stdout, "[%s:%s:%d] " fmt,                             \
                    __FILE__, __FUNCTION__, __LINE__, ##__VA_ARGS__);      \
    } while (0)

/**
 * Event handler for OpenXT wakeup requests.
 */
static void openxt_wakeup_requested(void * unused)
{
    char domain_path[256];
    int requested;

    (void)unused;

    // Determine the path that we'll be listening for wakeup events on.
    snprintf(domain_path, sizeof(domain_path),
             "/local/domain/%d", xen_domid);

    // Check to see if a wakeup has been r
    if(xenstore_read_int(domain_path, "wakeup-req", &requested) != 0)
    {
        ACPI_LOG("error: could not read the wakeup request node");
        return;
    }

    //... if it has, wake up the guest.
    if(requested)
    {

        // Clear the pending request...
        xenstore_write_int(domain_path, "wakeup-req", 0);

        // ... and issue our wakeup request!
        if(runstate_check(RUN_STATE_SUSPENDED)) {
            ACPI_LOG("QEMU: Waking device models from emulated S3.\n");
            qemu_system_wakeup_request(QEMU_WAKEUP_REASON_OTHER);
        }
    }
}

/**
 * Register a watcher to handle OpenXT wakeup requests.
 */
void acpi_openxt_register_wakeup_listener(void)
{
    char domain_path[256];

    // Ensure that our connection to the XenStore has been set up.
    xenstore_generic_init();

    // Determine the path that we'll be listening for wakeup events on.
    snprintf(domain_path, sizeof(domain_path),
             "/local/domain/%d", xen_domid);

    // Create an emtpy wakeup-request node, as a hint that we're listening
    // for wakeup requests.
    xenstore_write_int(domain_path, "wakeup-req", 0);

    //... and register an watch for the given node.
    if (xenstore_add_watch(domain_path, "wakeup-req",
          openxt_wakeup_requested, (void *) -1 )) {
        ACPI_LOG("error: failed to register watch for domain wakeup\n");
    }
}
